import logging

from gevent.pywsgi import WSGIServer
from mushroom.http import HttpError, HttpMethodNotAllowed, \
        HttpNotFound, HttpUnauthorized, HttpInternalServerError, \
        HttpNotImplemented, HttpRequest, HttpResponse, JsonResponse, \
        WebSocketTransport, PollTransport
from mushroom.rpc import Message, dummy_rpc_handler
from mushroom.session import session_id_generator, Session, SessionList, \
        SessionHandler


logger = logging.getLogger('mushroom.application')


class Application(object):

    def __init__(self, rpc_handler=None, session_handler=None):
        self.sessions = SessionList()
        self.sid_generator = session_id_generator()
        self.rpc_handler = rpc_handler or dummy_rpc_handler
        self.session_handler = session_handler or SessionHandler()

    def __call__(self, environ, start_response):
        try:
            try:
                request = HttpRequest(environ)
                response = self.request(request)
            except Exception as e:
                if isinstance(e, HttpError):
                    raise
                else:
                    logger.exception(e)
                    raise HttpInternalServerError
        except HttpError as e:
            response = HttpResponse(e.code, e.message, extra_headers=e.headers)
        # If the connection has switched to the WebSocket protocol no
        # response is expected. In this case the WSGI application simply
        # needs to returns None.
        if response:
            start_response(response.code, response.headers.items())
            return [response.content]

    def request(self, request):
        # / -> Authentication and connection bootstrapping
        if not request.path:
            return self.bootstrap(request)
        # /.*/
        if len(request.path) == 1:
            try:
                sid = request.path[0]
                session = self.sessions[sid]
            except KeyError:
                raise HttpNotFound
            return session.transport.handle_http_request(request, session)
        raise HttpNotFound

    def bootstrap(self, request):
        # Only allow POST requests for bootstrapping
        if request.method != 'POST':
            raise HttpMethodNotAllowed(['POST'])
        for transport in request.data['transports']:
            if transport == 'ws':
                return self.start_session(request, WebSocketTransport())
            if transport == 'poll':
                return self.start_session(request, PollTransport())
        # No suitable transport found
        raise HttpNotImplemented('No suitable transport found')

    def start_session(self, request, transport):
        session = Session(self.sid_generator.next(), transport, self.rpc_handler)
        if not self.session_handler.authenticate(session, request.data.get('auth', None)):
            raise HttpUnauthorized
        self.sessions.add(session)
        self.session_handler.connect(session)
        @transport.state.subscribe
        def state_listener(state):
            if state == 'DISCONNECTED':
                self.sessions.remove(session)
                self.session_handler.disconnect(session)
        return JsonResponse(transport.get_handshake_data(request, session))



try:
    import simplejson as json
except ImportError:
    import json

from django.http import (HttpResponse, JsonResponse,
        HttpResponseNotFound, HttpResponseNotAllowed,
        HttpResponseForbidden, HttpResponseBadRequest)

from mushroom.http import WebSocketTransport, PollTransport
from mushroom.session import session_id_generator, Session, SessionList, \
        SessionHandler, SessionHandlerAdapter
from mushroom.rpc.dispatcher import MethodDispatcher, dummy_rpc_handler
from mushroom.rpc.middleware import MiddlewareStack

from .middlewares import CloseOldConnectionsMiddleware


class HttpResponseNotImplemented(HttpResponse):
    status_code = 501


class MushroomView(object):
    csrf_exempt = True

    def __init__(self, rpc_handler=None, session_handler=None):
        self.sessions = SessionList()
        self.sid_generator = session_id_generator()
        self.rpc_handler = MiddlewareStack(
                self.get_middlewares(),
                rpc_handler or dummy_rpc_handler)
        self.session_handler = session_handler or SessionHandler()

    def get_middlewares(self):
        return [
           CloseOldConnectionsMiddleware
        ]

    def __call__(self, request, sid=None):
        if sid is None:
            response = self.bootstrap(request)
        else:
            try:
                session = self.sessions[sid]
            except KeyError:
                response = HttpResponseNotFound()
            else:
                response = session.transport.handle_http_request(request, session)
        if response is None:
            # Empty response to make Django happy
            response = HttpResponse()
        response['Access-Control-Allow-Origin'] = '*'
        return response

    def bootstrap(self, request):
        # Only allow POST requests for bootstrapping
        if request.method != 'POST':
            return HttpResponseNotAllowed(['POST'])
        try:
            data = json.loads(request.body)
        except ValueError:
            return HttpResponseBadRequest('Invalid JSON')
        if not isinstance(data, dict):
            return HttpResponseBadRequest('Invalid transport negotiation request')
        transports = data.get('transports', [])
        if not isinstance(transports, list):
            return HttpResponseBadRequest('Invalid transport negotiation request')
        for transport in data.get('transports', []):
            if transport == 'ws':
                return self.start_session(request, data, WebSocketTransport())
            if transport == 'poll':
                return self.start_session(request, data, PollTransport())
        # No suitable transport found
        return HttpResponseNotImplemented('No suitable transport found')

    def start_session(self, request, data, transport):
        session = Session(self.sid_generator.next(), transport, self.rpc_handler)
        if not self.session_handler.authenticate(session, data.get('auth', None)):
            return HttpResponseForbidden()
        self.sessions.add(session)
        self.session_handler.connect(session)
        @transport.state.subscribe
        def state_listener(state):
            if state == 'DISCONNECTED':
                self.sessions.remove(session)
                self.session_handler.disconnect(session)
        return JsonResponse(transport.get_handshake_data(request, session))


class SimpleMushroomView(MushroomView):

    def __init__(self, *args, **kwargs):
        if 'rpc_handler' not in kwargs:
            kwargs['rpc_handler'] = MethodDispatcher(self, 'rpc_')
        if 'session_handler' not in kwargs:
            kwargs['session_handler'] = SessionHandlerAdapter(self, 'session_')
        super(SimpleMushroomView, self).__init__(*args, **kwargs)



from datetime import datetime
import functools
import logging
import sys
from time import time

import six

from .messages import Notification
from .messages import Request


logger = logging.getLogger('mushroom.rpc.middleware')


class Middleware(object):

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)


class MiddlewareStack(object):

    def __init__(self, middlewares, get_response):
        self.get_response = get_response
        for m in reversed(middlewares):
            self.get_response = m(self.get_response)

    def __call__(self, request):
        return self.get_response(request)


class SlowMethodLogMiddleware(Middleware):
    '''
    Default thresholds:
    (10, 'error'), (1, 'warning')
    '''

    def __init__(self, get_response, thresholds=None):
        super(SlowMethodLogMiddleware, self).__init__(get_response)
        if thresholds is None:
            self.thresholds = [(10, 'error'), (1, 'warning')]
        else:
            self.thresholds = list(thresholds)
            self.thresholds.sort(reverse=True)

    def __call__(self, request):
        request.start_time = time()
        try:
            return super(SlowMethodLogMiddleware, self).__call__(request)
            self.log_if_slow(request, 'Slow method "%s" returned normally after %.3fs')
        except:
            self.log_if_slow(request, 'Slow method "%s" returned an error after %.3fs')
            raise

    def log_if_slow(self, request, message):
        duration = time() - request.start_time
        for threshold, log_level in self.thresholds:
            if duration > threshold:
                log_args = (message, request.method, duration)
                log_method = getattr(logger, log_level)
                log_method(*log_args)
                break


class StatisticsMiddleware(Middleware):

    def __call__(self, request):
        request.session.last_activity = datetime.now()
        request.session.message_count = getattr(request.session, 'message_count', 0) + 1
        return super(StatisticsMiddleware, self).__call__(request)


class RunningMethod(object):

    def __init__(self, method_name, data):
        self.name = method_name
        self.data = data
        self.start_time = time()

    @property
    def elapsed_seconds(self):
        return time() - self.start_time


class RunningMethodsMiddleware(Middleware):

    def __call__(self, request):
        request.session.running_methods = getattr(request.session, 'running_methods', [])
        running_method = RunningMethod(request.method, request.data)
        request.session.running_methods.append(running_method)
        try:
            return super(RunningMethodsMiddleware, self).__call__(request)
        finally:
            request.session.running_methods.remove(running_method)



from functools import partial

import six

from .exceptions import MethodNotFound


class MethodDispatcher(object):
    '''
    Dispatcher implementation that calls methods on an object with a
    specific prefix and/or suffix. This makes it possible to define
    objects that provides a set of methods which can then be called by
    the client.

    Note: Using an empty prefix, ``_`` or ``__`` is highly discouraged as
    it allows the client to call methods like methods like ``__del__``.
    The same holds true for the suffix. If you really want to dispatch
    methods without a prefix or suffix it is a good idea to write a
    custom dispatcher that implements some checks for this.
    '''

    def __init__(self, obj, prefix='rpc_', suffix=''):
        '''
        Constructor for the method dispatcher.

        :param obj: object instance which is used for the method lookup
        :param prefix: string prefix which is prepended to the method name
            when looking up the method
        :param suffix: string suffix which is appended to the method name
            when looking up the method
        '''
        self.obj = obj
        self.prefix = prefix
        self.suffix = suffix

    def __call__(self, request):
        '''
        The `Engine` calls the request handler like it was a function
        that takes the request as sole argument and returns the response.
        This function implements the adapter for this interface and makes
        it possible to use this class as `handler` for the `Engine`.

        :param request: request object
        '''
        method_name = self.prefix + request.method + self.suffix
        try:
            method = getattr(self.obj, method_name)
        except AttributeError:
            raise MethodNotFound(method_name)
        return method(request)


class RegistryDispatcher(object):
    '''
    Dispatcher implementation that uses a registry aproach.
    '''

    def __init__(self):
        self.methods = {}

    def register(self, *args):
        '''
        This method can be used in three ways:

        1.) @dispatcher.register('ping')
            def ping(request):
                return 'pong'

        2.) @dispatcher.register
            def ping(request):
                return 'pong'

        3.) def ping(request):
                return 'pong'
            dispatcher.register('ping', ping)
        '''
        if len(args) == 0:
            return self.register
        if len(args) == 1:
            if callable(args[0]):
                function = args[0]
                return self.register(function.__name__, function)
            elif isinstance(args[0], six.string_types):
                name = args[0]
                return lambda function: self.register(name, function)
            else:
                raise TypeError('First argument must be either callable or string')
        if len(args) == 2:
            name = args[0]
            function = args[1]
            if name in self.methods:
                raise TypeError('Method with name %r already registered.' % name)
            if hasattr(function, 'as_function'):
                function = function.as_function()
            self.methods[name] = function
            return function
        raise TypeError('%s() takes 0 to 2 arguments (%d given)' % (
                self.register.__name__, len(args)))

    def __call__(self, request):
        '''
        The `Engine` calls the request handler like it was a function
        that takes the request as sole argument and returns the response.
        This function implements the adapter for this interface and makes
        it possible to use this class as `handler` for the `Engine`.

        :param request: request object
        '''
        method_name = request.method
        try:
            method = self.methods[method_name]
        except KeyError:
            raise MethodNotFound(method_name)
        return method(request)


def dummy_rpc_handler(request):
    '''
    Dummy RPC handler that raises a MethodNotFound exception for
    all calls. This is useful for applications that do not need do
    receive any data from the client but only publish data.
    '''
    raise MethodNotFound(request.method)



#!/usr/bin/env python

import example_pythonpath
from example_utils import ExampleServer, ExampleStaticFile

import gevent
from gevent import subprocess

class WebexecServer(ExampleServer):
    urls = [
        ('/', ExampleStaticFile('webexec.html')),
    ] + ExampleServer.urls

    def server_init(self):
        gevent.spawn(self.exec_subprocess)

    def exec_subprocess(self):
        proc = subprocess.Popen(['ping', 'localhost'], stdout=subprocess.PIPE)
        while True:
            line = proc.stdout.readline().rstrip()
            if line is None:
                break
            self.sessions.notify('stdout', line)


if __name__ == '__main__':
    WebexecServer.main()



import functools


class Signal(object):

    def __init__(self):
        self.receivers = []

    def connect(self, receiver):
        self.receivers.append(receiver)

    def disconnect(self, receiver):
        self.receivers.remove(receiver)

    def send(self, *args, **kwargs):
        for receiver in self.receivers:
            receiver(*args, **kwargs)


pre_message = Signal()
post_message = Signal()


def message_handler(handler):
    '''Decorator for message handler which makes sure that a pre_message
    and post_message signal is sent before and after the message. This
    provides a hook for the application to open/close database connections
    and perform other clean up tasks required on a per request basis.'''
    @functools.wraps(handler)
    def wrapper():
        pre_message.send()
        try:
            handler()
        finally:
            post_message.send()
    return wrapper



#!/usr/bin/env python

import example_pythonpath
from example_utils import ExampleServer, ExampleStaticFile


class ChatServer(ExampleServer):
    urls = [
        ('/', ExampleStaticFile('chat.html')),
    ] + ExampleServer.urls

    def rpc_message(self, request):
        self.sessions.notify('message', request.data)


if __name__ == '__main__':
    ChatServer.main()



import unittest

from mushroom.rpc.middleware import Middleware, MiddlewareStack
from mushroom.rpc.exceptions import RpcError
from mushroom.rpc.exceptions import RequestException
from mushroom.rpc.dispatcher import MethodDispatcher


class MockRequest(object):

    def __init__(self, method, data=None):
        self.method = method
        self.data = data


class MyError(RuntimeError):
    pass


class ErrorMiddleware(Middleware):

    def __call__(self, request):
        try:
            return super(ErrorMiddleware, self).__call__(request)
        except MyError as e:
            raise RequestException(e.message)


count = 0;


class CountMiddleware(Middleware):

    def __call__(self, request):
        global count
        count += 1
        try:
            return super(CountMiddleware, self).__call__(request)
        finally:
            count -= 1


class RpcMethods(object):

    def rpc_fail(self, request):
        raise MyError('Fail!')

    def rpc_count(self, request):
        return count


class MiddlewareTestCase(unittest.TestCase):

    def setUp(self):
        self.rpc_handler = MiddlewareStack([
            ErrorMiddleware,
            CountMiddleware
        ], MethodDispatcher(RpcMethods()))

    def test_error(self):
        self.assertRaises(
                RequestException,
                self.rpc_handler,
                MockRequest('fail'))
        self.assertEquals(count, 0)

    def test_count(self):
        self.assertEquals(self.rpc_handler(MockRequest('count', None)), 1)
        self.assertEquals(count, 0)



# This file adds the mushroom project root to the PYTHONPATH
# making it possible to run the examples directly from the
# checked out mushroom repository without having to install
# it first.

import os
import sys

EXAMPLES_DIR = os.path.dirname(os.path.abspath(__file__))
PROJECT_ROOT = os.path.dirname(EXAMPLES_DIR)

os.sys.path[1:1] = [PROJECT_ROOT]



#!/usr/bin/env python

import example_pythonpath
from example_utils import ExampleServer, ExampleStaticFile


class PingServer(ExampleServer):
    urls = [
        ('/', ExampleStaticFile('ping.html')),
    ] + ExampleServer.urls

    def rpc_ping(self, request):
        return 'pong'


if __name__ == '__main__':
    PingServer.main()
